/*
 * ValueHolderInterface.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _VALUE_HOLDER_INTERFACE_HPP_
#define _VALUE_HOLDER_INTERFACE_HPP_

#include <abstraction/Value.hpp>

namespace abstraction {

template < class Type >
class ValueHolderInterface : public Value {
	ParamQualifiers::ParamQualifierSet m_paramQualifierSet;
	bool m_isTemporary;

protected:
	ValueHolderInterface ( const std::shared_ptr < abstraction::OperationAbstraction > & ref, ParamQualifiers::ParamQualifierSet paramQualifierSet, bool isTemporary ) : Value ( ref ), m_paramQualifierSet ( paramQualifierSet ), m_isTemporary ( isTemporary ) {
		if ( this->isLvalueRef ( ) && this->isTemporary ( ) )
			throw std::domain_error ( "Lvalue references cannot be a temporarie." );
		if ( this->isLvalueRef ( ) && this->isRvalueRef ( ) )
			throw std::domain_error ( "A reference cannot be both, rvalue and lvalue." );
	}

public:
	virtual Type && getValue ( ) const = 0;

	ParamQualifiers::ParamQualifierSet getParamQualifiersSet ( ) const {
		return m_paramQualifierSet;
	}

	bool isConst ( ) const {
		return m_paramQualifierSet && ParamQualifiers::ParamQualifierSet::CONST;
	}

	bool isRef ( ) const {
		return isRvalueRef ( ) || isLvalueRef ( );
	}

	bool isRvalueRef ( ) const {
		return m_paramQualifierSet && ParamQualifiers::ParamQualifierSet::RREF;
	}

	bool isLvalueRef ( ) const {
		return m_paramQualifierSet && ParamQualifiers::ParamQualifierSet::LREF;
	}

	bool isTemporary ( ) const {
		return m_isTemporary;
	}

};

template < class ParamType >
ParamType retrieveValue ( const std::shared_ptr < abstraction::Value > & param, bool move ) {
	(void) move;
	using Type = std::decay_t < ParamType >;

	std::shared_ptr < ValueHolderInterface < Type > > interface = std::dynamic_pointer_cast < ValueHolderInterface < Type > > ( param->getProxyAbstraction ( ) );
	if ( ! interface )
		throw std::invalid_argument ( "Abstraction does not provide value of type " + ext::to_string < ParamType > ( ) + " but " + param->getType ( ) + "." );

	if constexpr ( std::is_rvalue_reference_v < ParamType > ) {
		if ( move || interface->isTemporary ( ) )
			return std::move ( interface->getValue ( ) );
		else
			throw std::domain_error ( "Cannot bind without move" );
	} else if constexpr ( std::is_lvalue_reference_v < ParamType > && std::is_const_v < std::remove_reference_t < ParamType > > ) {
		Type && res = interface->getValue ( );
		return res;
	} else if constexpr ( std::is_lvalue_reference_v < ParamType > ) {
		if ( interface->isTemporary ( ) && ! interface->isRef ( ) )
			throw std::domain_error ( "Cannot bind temporary to non-const reference" );
		Type && res = interface->getValue ( );
		return res;
	} else if constexpr ( std::is_copy_constructible_v < Type > && std::is_move_constructible_v < Type > ) {
		if ( ! interface->isConst ( ) && ( move || interface->isTemporary ( ) ) )
			return std::move ( interface->getValue ( ) );
		else {
			const Type & res = interface->getValue ( );
			return res;
		}
	} else if constexpr ( std::is_copy_constructible_v < Type > && ! std::is_move_constructible_v < Type > ) {
		if ( ! interface->isConst ( ) && ( move || interface->isTemporary ( ) ) )
			throw std::domain_error ( "Value not move constructible" );
		else {
			Type && res = interface->getValue ( );
			return res;
		}
	} else if constexpr ( ! std::is_copy_constructible_v < Type > && std::is_move_constructible_v < Type > ) {
		if ( ! interface->isConst ( ) && ( move || interface->isTemporary ( ) ) )
			return std::move ( interface->getValue ( ) );
		else
			throw std::domain_error ( "Value not copy constructible" );
	} else { // ! std::is_copy_constructible_v < Type > && ! std::is_move_constructible_v < Type >
		if ( ! interface->isConst ( ) && move )
			throw std::domain_error ( "Value not move constructible" );
		else
			throw std::domain_error ( "Value not copy constructible" );
	}
}

} /* namespace abstraction */

#endif /* _VALUE_HOLDER_INTERFACE_HPP_ */
